depends = ('ITKCommon',)
templates = (
    ('DIMS', 'ITK_WRAP_IMAGE_DIMS', 'DIMS'),
    ('USIGN_INTS', 'WRAP_ITK_USIGN_INT', 'USIGN_INTS'),
    ('SIGN_INTS', 'WRAP_ITK_SIGN_INT', 'SIGN_INTS'),
    ('REALS', 'WRAP_ITK_REAL', 'REALS'),
    ('VECTOR_REALS', 'WRAP_ITK_VECTOR_REAL', 'VECTOR_REALS'),
    ('COV_VECTOR_REALS', 'WRAP_ITK_COV_VECTOR_REAL', 'COV_VECTOR_REALS'),
    ('RGBS', 'WRAP_ITK_RGB', 'RGBS'),
    ('RGBAS', 'WRAP_ITK_RGBA', 'RGBAS'),
    ('COMPLEX_REALS', 'WRAP_ITK_COMPLEX_REAL', 'COMPLEX_REALS'),
    ('INTS', 'WRAP_ITK_INT', 'INTS'),
    ('SCALARS', 'WRAP_ITK_SCALAR', 'SCALARS'),
    ('VECTORS', 'WRAP_ITK_VECTOR', 'VECTORS'),
    ('COLORS', 'WRAP_ITK_COLOR', 'COLORS'),
    ('ALL_TYPES', 'WRAP_ITK_ALL_TYPE', 'ALL_TYPES'),
)
